/* rbldnsd_dnvset.c.
 * Zone type which consists of a set of (possible wildcarded)
 * domain names together with (A,TXT) result for each.
 */

#include <syslog.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "rbldnsd.h"
#include "dns.h"
#include "mempool.h"
#include "sstr.h"

struct entry {
  unsigned char *dn; /* mp-allocated */
  ip4addr_t r_a;
  char *r_txt;  /* sstr */
};

/* zone data and operations is the same as in rbldnsd_dnset,
 * except of instead array of just domain names we store array
 * of structures for each entry.  Also, we allow duplicated entries. */

struct zonedata {
  int n[2]; /* number of entries */
  int a[2]; /* entries allocated (used only when loading) */
  struct entry *e[2]; /* entries: plain and wildcard */
  unsigned maxlab[2]; /* max level of labels */
  ip4addr_t r_a; /* result: address */
  char *r_txt; /* result: text */
  struct mempool mp; /* mempool for domain names */
};

/* indexes */
#define EP 0
#define EW 1

static void dnvset_freeent(struct entry *e, int n) {
  while(n--)
    if (e[n].r_txt) sstrfree(e[n].r_txt);
  free(e);
}

void dnvset_free(struct zonedata *z) {
  if (z) {
    mp_free(&z->mp);
    if (z->r_txt) sstrfree(z->r_txt);
    if (z->e[EP]) dnvset_freeent(z->e[EP], z->n[EP]);
    if (z->e[EW]) dnvset_freeent(z->e[EW], z->n[EW]);
    free(z);
  }
}

static int
dnvset_addent(struct zonedata *z, int idx, unsigned char *dn,
             ip4addr_t r_a, const char *r_txt) {
  struct entry *e = z->e[idx];
  unsigned labels;
  if (z->n[idx] >= z->a[idx]) {
    z->a[idx] = z->a[idx] ? z->a[idx] << 1 : 32;
    e = z->e[idx] = (struct entry *)erealloc(e, z->a[idx] * sizeof(*e));
    if (!e) return 0;
  }
  e += z->n[idx]++;
  e->r_a = r_a ? r_a : z->r_a;
  if (!*r_txt) e->r_txt = NULL;
  else if (!(e->r_txt = esstrdup(r_txt))) {
    e->dn = NULL;
    return 0;
  }
  if (!(e->dn = (unsigned char*)mp_estrdup(&z->mp, dn)))
    return 0;
  labels = dns_dnlabels(dn);
  if (z->maxlab[idx] < labels) z->maxlab[idx] = labels;
  return 1;
}

static int
dnvset_loadfile(FILE *f, struct zonedata *z) {
  char buf[512];
  int lineno = 0;
  unsigned l;
  char *p;
  ip4addr_t a;
  unsigned char dn[DNS_MAXDN];

  while(fgets(buf, sizeof(buf), f)) {
    ++lineno;
    l = strlen(buf);
    if (l && buf[l-1] == '\n') buf[--l] = '\0';
    if (!l || buf[0] == '#') continue;
    if (lineno == 1 && buf[0] == ':') {
      if (!addrtxt(buf, &a, &p)) {
        zlog(LOG_WARNING, 1, "invalid default entry");
	continue;
      }
      if (a) z->r_a = a;
      if (*p && !(z->r_txt = esstrdup(p)))
        return 0;
      continue;
    }
    for(p = buf; *p; ++p)
      if (*p == ' ' || *p == '\t' || *p == '#') break;
    if (*p) *p++ = '\0';
    if (!dns_ptodn(buf, dn, DNS_MAXDN)) {
      zlog(LOG_WARNING, lineno, "invalid domain name `%s'", buf);
      continue;
    }
    dns_dntol(dn, dn);
    while(*p == ' ' || *p == '\t') ++p;
    if (*p && !addrtxt(p, &a, &p)) {
      zlog(LOG_WARNING, lineno, "invalid value");
      continue;
    }
    if (dn[0] == 1 && dn[1] == '*') {
      if (!dnvset_addent(z, EW, dn + 2, a, p)) return 0;
    }
    else {
      if (!dnvset_addent(z, EP, dn, a, p)) return 0;
    }
  }

  return 1;
}

static int dnvset_cmpent(const struct entry *a, const struct entry *b) {
  return strcmp(a->dn, b->dn);
}

struct zonedata *dnvset_load(FILE *f) {
  int r;
  struct zonedata *z = (struct zonedata *)emalloc(sizeof(*z));
  if (!z) return NULL;
  memset(z, 0, sizeof(*z));
  z->r_a = R_A_DEFAULT;
  if (!dnvset_loadfile(f, z)) {
    dnvset_free(z);
    return NULL;
  }
  for(r = 0; r < 2; ++r) {
    if (!z->n[r]) continue;
    if (z->a[r] != z->n[r])
      z->e[r] = (struct entry *)
        realloc(z->e[r], z->n[r] * sizeof(struct entry));
    qsort(z->e[r], z->n[r], sizeof(struct entry),
          (int(*)(const void*, const void*))dnvset_cmpent);
  }
  return z;
}

static const struct entry *
dnvset_find(const struct entry *e, int b, const unsigned char *q) {
  int a = 0, m, r;
  while(a <= b) {
    m = (a + b) >> 1;
    r = strcmp(e[m].dn, q);
    if (!r) {
      while(m && strcmp(e[m-1].dn, q) == 0) --m;
      return e + m;
    }
    else if (r < 0) a = m + 1;
    else b = m - 1;
  }
  return NULL;
}

int dnvset_query(const struct zonedata *const z, struct dnspacket *p,
                 const unsigned char *const query, unsigned qflg)
{
  const unsigned char *dn = query;
  unsigned labels = dns_dnlabels(dn);
  const struct entry *e, *i, *t;
  if (labels > z->maxlab[EP] || !z->n[EP] ||
      !(e = dnvset_find(z->e[EP], z->n[EP] - 1, dn))) {
    /* try wildcard */
    while(labels > z->maxlab[EW])
      --labels, dn += 1 + *dn;
    for(;;) {
      if (!*dn) return 0;
      if ((e = dnvset_find(z->e[EW], z->n[EW]-1, dn)) != NULL) break;
      dn += 1 + *dn;
    }
    t = z->e[EW] + z->n[EW];
  }
  else
    t = z->e[EP] + z->n[EP];
  if (qflg & QUERY_A) {
    i = e;
    do addrec_a(p, i->r_a);
    while(++i < t && strcmp(i->dn, dn) == 0);
  }
  if (qflg & QUERY_TXT) {
    char name[DNS_MAXDOMAIN+1];
    const char *txt;
    dns_dntop(query, name, sizeof(name));
    i = e;
    do
      if ((txt = i->r_txt ? i->r_txt : z->r_txt) != NULL)
        addrec_txt(p, txt, name);
    while(++i < t && strcmp(i->dn, dn) == 0);
  }
  return 1;
}
